/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2005  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"
#include "types.h"
#include "logging.h"

#include <string.h>

#include "fs.h"
#include "keyboard.h"
#include "m6809.h"
#include "machine.h"
#include "pia.h"
#include "sam.h"
#include "tape.h"
#include "vdg.h"
#include "wd2797.h"
#include "xroar.h"

/* machine indicates which machine architecture is selected.
 * machine_keyboard tracks which keyboard is in use (good to be able to
 * set it separately). */
int machine, machine_keyboard;
int dragondos_enabled;
uint8_t ram0[0x8000];
uint8_t ram1[0x8000];
uint8_t rom0[0x8000];
uint8_t rom1[0x8000];
int_fast32_t brk_csrdon, brk_bitin;

/* These list keyboard matrix {column,row} points for keys by ASCII value.
 * Individual keyboard modules are responsible for mapping host keys to
 * ASCII codes.  {8,8} means there is no key corresponding to that code.
 * Code 0 maps to the shift key. */
static Keyboard dragon_keyboard = {
	{7,6}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, /* 0 - 7 */
	{5,5}, {6,5}, {4,5}, {8,8}, {1,6}, {0,6}, {8,8}, {8,8}, /* 8 - 15 */
	{8,8}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, /* 16 - 23 */
	{8,8}, {8,8}, {8,8}, {2,6}, {8,8}, {8,8}, {8,8}, {8,8}, /* 24 - 31 */
	{7,5}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, /* 32 - 39 */
	{8,8}, {8,8}, {8,8}, {8,8}, {4,1}, {5,1}, {6,1}, {7,1}, /* 40 - 47 */
	{0,0}, {1,0}, {2,0}, {3,0}, {4,0}, {5,0}, {6,0}, {7,0}, /* 48 - 55 */
	{0,1}, {1,1}, {2,1}, {3,1}, {8,8}, {8,8}, {8,8}, {8,8}, /* 56 - 63 */
	{0,2}, {1,2}, {2,2}, {3,2}, {4,2}, {5,2}, {6,2}, {7,2}, /* 64 - 71 */
	{0,3}, {1,3}, {2,3}, {3,3}, {4,3}, {5,3}, {6,3}, {7,3}, /* 72 - 79 */
	{0,4}, {1,4}, {2,4}, {3,4}, {4,4}, {5,4}, {6,4}, {7,4}, /* 80 - 87 */
	{0,5}, {1,5}, {2,5}, {8,8}, {8,8}, {8,8}, {3,5}, {8,8}, /* 88 - 95 */
	{1,6}, {1,2}, {2,2}, {3,2}, {4,2}, {5,2}, {6,2}, {7,2}, /* 96 - 103 */
	{0,3}, {1,3}, {2,3}, {3,3}, {4,3}, {5,3}, {6,3}, {7,3}, /* 104 - 111 */
	{0,4}, {1,4}, {2,4}, {3,4}, {4,4}, {5,4}, {6,4}, {7,4}, /* 112 - 119 */
	{0,5}, {1,5}, {2,5}, {8,8}, {8,8}, {8,8}, {8,8}, {5,5}, /* 120 - 127 */
};
static Keyboard coco_keyboard = {
	{7,6}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, /* 0 - 7 */
	{5,3}, {6,3}, {4,3}, {8,8}, {1,6}, {0,6}, {8,8}, {8,8}, /* 8 - 15 */
	{8,8}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, /* 16 - 23 */
	{8,8}, {8,8}, {8,8}, {2,6}, {8,8}, {8,8}, {8,8}, {8,8}, /* 24 - 31 */
	{7,3}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, {8,8}, /* 32 - 39 */
	{8,8}, {8,8}, {8,8}, {8,8}, {4,5}, {2,5}, {6,5}, {7,5}, /* 40 - 47 */
	{0,4}, {1,4}, {2,4}, {3,4}, {4,4}, {5,4}, {6,4}, {7,4}, /* 48 - 55 */
	{0,5}, {1,5}, {2,5}, {3,5}, {8,8}, {5,5}, {8,8}, {8,8}, /* 56 - 63 */
	{0,0}, {1,0}, {2,0}, {3,0}, {4,0}, {5,0}, {6,0}, {7,0}, /* 64 - 71 */
	{0,1}, {1,1}, {2,1}, {3,1}, {4,1}, {5,1}, {6,1}, {7,1}, /* 72 - 79 */
	{0,2}, {1,2}, {2,2}, {3,2}, {4,2}, {5,2}, {6,2}, {7,2}, /* 80 - 87 */
	{0,3}, {1,3}, {2,3}, {0,0}, {8,8}, {8,8}, {3,3}, {8,8}, /* 88 - 95 */
	{1,6}, {1,0}, {2,0}, {3,0}, {4,0}, {5,0}, {6,0}, {7,0}, /* 96 - 103 */
	{0,1}, {1,1}, {2,1}, {3,1}, {4,1}, {5,1}, {6,1}, {7,1}, /* 104 - 111 */
	{0,2}, {1,2}, {2,2}, {3,2}, {4,2}, {5,2}, {6,2}, {7,2}, /* 112 - 119 */
	{0,3}, {1,3}, {2,3}, {8,8}, {8,8}, {8,8}, {8,8}, {5,3}, /* 120 - 127 */
};
/* The selected keyboard is copied into here: */
Keyboard keyboard;

/* When a config file exists, these lists will disappear and just become
 * single configured filenames */
char *d32_roms[] = { "dragon32", "d32rom", "d32", "dragon", NULL };
char *coco_roms[] = { "coco", "tandyrom", "coco_pa", "cocodisk", NULL };
char *d64_1_roms[] = { "d64rom1", "d64_1", "dragon64", "d64", "dragrom", NULL };
char *d64_2_roms[] = { "d64rom2", "d64_2", NULL };
char *dragondos_roms[] = { "sdose6", "ddos10", NULL };

static int load_rom(char **rom_list, uint8_t *dest, size_t max_size);

void machine_init(void) {
	machine_set(DRAGON64);
	dragondos_enabled = 1;
	sam_init();
	pia_init();
	wd2797_init();
	m6809_init();
	vdg_init();
	tape_init();
}

void machine_reset(int hard) {
	if (hard) {
		memset(ram0, 0, sizeof(ram0));
		memset(ram1, 0, sizeof(ram1));
		memset(rom0, 0, sizeof(rom0));
		memset(rom0, 0, sizeof(rom1));
		if (IS_COCO) {
			brk_csrdon = 0xa77c; brk_bitin = 0xa755;
			dragondos_enabled = 0;
			load_rom(coco_roms, rom0, 0x4000);
		} else {
			brk_csrdon = 0xbde7; brk_bitin = 0xbda5;
			if (IS_DRAGON64) {
				if (!load_rom(d64_1_roms, rom0, 0x4000))
					load_rom(d32_roms, rom0, 0x4000);
				load_rom(d64_2_roms, rom1, 0x4000);
			} else {
				load_rom(d32_roms, rom0, 0x4000);
			}
			if (dragondos_enabled)
				if (!load_rom(dragondos_roms, rom0+0x4000, 0x3f00))
					dragondos_enabled = 0;
			/* Accounts for D64 ROM select not affecting the
			 * top 16K */
			memcpy(rom1+0x4000, rom0+0x4000, 0x4000);
		}
	}
	pia_reset();
	if (IS_DRAGON64) {
		PIA_1B.port_input |= (1<<2);
		PIA_UPDATE_OUTPUT(PIA_1B);
	}
	wd2797_reset();
	sam_reset();
	m6809_reset();
	vdg_reset();
	tape_reset();
}

/* Selects a machine architecture, loads the relevant ROMs and resets */
void machine_set(int mode) {
	if (mode >= NUM_MACHINES) mode = 0;
	machine = mode;
	if (IS_DRAGON) {
		dragondos_enabled = 1;
		machine_set_keyboard(DRAGON_KEYBOARD);
	} else {
		machine_set_keyboard(COCO_KEYBOARD);
	}
	xroar_reset(RESET_HARD);
}

/* Setting keyboard takes effect immediately */
void machine_set_keyboard(int mode) {
	if (mode >= NUM_KEYBOARDS) mode = 0;
	machine_keyboard = mode;
	switch (mode) {
		default:
		case DRAGON_KEYBOARD:
			memcpy(keyboard, dragon_keyboard, sizeof(keyboard));
			break;
		case COCO_KEYBOARD:
			memcpy(keyboard, coco_keyboard, sizeof(keyboard));
			break;
	}
}

static int load_rom(char **rom_list, uint8_t *dest, size_t max_size) {
	FS_FILE fd;
	char filename[sizeof(ROMPATH)+14];
	int i;
	/* First check for any file in the list existing in ROMPATH */
	for (i = 0, fd = -1; fd == -1 && rom_list[i]; i++) {
		strncpy(filename, ROMPATH, sizeof(ROMPATH));
		strcat(filename, rom_list[i]);
		strcat(filename, ".rom");
		LOG_DEBUG(4,"Checking for %s...\n", filename);
		if ((fd = fs_open(filename, FS_READ)) == -1) {
			strncpy(filename, ROMPATH, sizeof(ROMPATH));
			strcat(filename, rom_list[i]);
			strcat(filename, ".dgn");
			LOG_DEBUG(4,"Checking for %s...\n", filename);
			fd = fs_open(filename, FS_READ);
		}
	}
	/* Then check for the same filenames in current directory */
	for (i = 0; fd == -1 && rom_list[i]; i++) {
		strcpy(filename, rom_list[i]);
		strcat(filename, ".rom");
		LOG_DEBUG(4,"Checking for %s...\n", filename);     
		if ((fd = fs_open(filename, FS_READ)) == -1) {     
			strcpy(filename, rom_list[i]);
			strcat(filename, ".dgn");                    
			LOG_DEBUG(4,"Checking for %s...\n", filename);
			fd = fs_open(filename, FS_READ);
		}
	}
	if (fd == -1) return 0;
	LOG_DEBUG(4,"Loading %s...\n", filename);
	if (fs_read(fd, dest, 16) < 0) {
		fs_close(fd);
		return 0;
	}
	if (dest[0] == 0xa1) {
		fs_read(fd, dest, max_size);
	} else {
		fs_read(fd, dest+16, max_size-16);
	}
	fs_close(fd);
	return 1;
}
